package dicewars.model;

import dicewars.util.Directions;
import dicewars.util.Point;

import java.util.ArrayList;
import java.util.List;

import static java.lang.Math.*;
import static java.lang.System.out;

/*
 *   A Hexagon suitable to use in a Hex grid (for skewed game grids)
 *
 *   The hex grid is (hex at (0,0) is at top left)
 *
 *   (0, 0) ---------> xh
 *     \
 *      \
 *       \
 *        V    yh (skewed y axis)
 *
 *
 *    *** Nothing to do here! ***
 *
 */
public class Hex {

    public static final double SQRT3 = sqrt(3);  // Pre-calculated
    private final int xh;         // Coordinate for horizontal axis
    private final int yh;         // Coordinate for skew vertical axis, pointing down

    public Hex(int xh, int yh) {
        this.xh = xh;
        this.yh = yh;
    }

    // Convert cartesian coordinates to hex coordinates
    // Radius is from center to any corner
    public Hex(Point point, double radius) {
        double x = (point.x * SQRT3/3 - point.y / 3) / radius;
        double y = point.y * 2/3 / radius;
        /*
            If doing just this

            this.xh = (int) round(x);
            this.yh = (int) round(y);

            We may get the wrong xh and yh due to rounding errors
            So check with any neighbour if closer to point
         */
        int xh = (int) round(x);
        int yh = (int) round(y);
        Hex tmp = new Hex(xh, yh);    // Assume this is the best
        List<Hex> nbs = tmp.getNeighbours();
        for (Hex h : nbs) {
            double d1 = h.toPoint(radius).distance(point);
            double d2 = tmp.toPoint(radius).distance(point);
            if (d1 < d2) {
                tmp = h;
            }
        }
        this.xh = tmp.xh;
        this.yh = tmp.yh;

    }

    // Convert hex coordinates to cartesian coordinates i.e.
    // get cartesian center point for some hex
    // For hex = (0,0) will get cartesian point (0,0)
    // Cartesian y-axis is screen y-axis (i.e. pointing down)
    public Point toPoint(double radius) {
        double x = radius * SQRT3 * (xh + (double) yh / 2);
        double y = radius * 3 / 2 * yh;
        return new Point(x, y);
    }

    /*
        The numbering of hexagon corners
              4
           3     5
           2     0
              1
     */
    public List<Point> getCorners(double radius) {
        Point center = toPoint(radius);
        List<Point> pts = new ArrayList<>();
        for (int i = 0; i < 6; i++) {
            double deg = 60 * i + 30;
            double rad = PI / 180 * deg;
            pts.add(new Point(center.x + radius * cos(rad), center.y + radius * sin(rad)));
        }
        return pts;
    }

    public List<Hex> getNeighbours() {
        List<Hex> neighbours = new ArrayList<>();
        for (Directions d : Directions.values()) {
            neighbours.add(new Hex(xh + d.dX, yh + d.dY));
        }
        return neighbours;
    }

    public int getXh() {
        return xh;
    }

    public int getYh() {
        return yh;
    }

    @Override
    public boolean equals(Object o) {   // Must have because we store hex's in collections!
        if (this == o) {
            return true;
        }
        if (!(o instanceof Hex)) {
            return false;
        }
        Hex hex = (Hex) o;
        if (xh != hex.xh) return false;
        return yh == hex.yh;
    }

    @Override
    public int hashCode() {     // Must have because we store hex's in maps!
        int result = xh;
        result = 31 * result + yh;
        return result;
    }

    @Override
    public String toString() {
        return "Hex{" + xh + ", " + yh + '}';
    }


}
